# Build Your Own Programming Language 

<a href="https://www.packtpub.com/programming/build-your-own-programming-language?utm_source=github&utm_medium=repository&utm_campaign=9781800204805"><img src="https://static.packt-cdn.com/products/9781800204805/cover/smaller" alt="Build Your Own Programming Language " height="256px" align="right"></a>

This is the code repository for [Build Your Own Programming Language ](https://www.packtpub.com/programming/build-your-own-programming-language?utm_source=github&utm_medium=repository&utm_campaign=9781800204805), published by Packt.

**A programmer's guide to designing compilers, interpreters, and DSLs for solving modern computing problems**

## What is this book about?
The need for different types of computer languages is growing rapidly and developers prefer creating domain-specific languages for solving specific application domain problems. Building your own programming language has its advantages. It can be your antidote to the ever-increasing size and complexity of software. However, creating a custom language isn't easy. 

This book covers the following exciting features:
- Perform requirements analysis for the new language and design language syntax and semantics.
- Write lexical and context-free grammar rules for common expressions and control structures.
- Develop a scanner that reads source code and generate a parser that checks syntax.
- Build key data structures in a compiler and use your compiler to build a syntax-coloring code editor.
- Implement a bytecode interpreter and run bytecode generated by your compiler.
- Write tree traversals that insert information into the syntax tree.
- Implement garbage collection in your language.

If you feel this book is for you, get your [copy](https://www.amazon.com/dp/1800204809) today!


## Instructions and Navigations
All of the code is organized into folders. For example, Chapter02.

The code will look like the following:
```
procedure main(argv)
    simple := simple()
    yyin := open(argv[1])
    while i := yylex() do
        write(yytext, ": ", i)
end
```

**Following is what you need for this book:**
This book is for software developers interested in the idea of inventing their own language or developing a domain-specific language. Computer science students taking compiler construction courses will also find this book highly useful as a practical guide to language implementation to supplement more theoretical textbooks. Intermediate-level knowledge and experience working with a high-level language such as Java or the C++ language are expected to help you get the most out of this book.

With the following software and hardware list you can run all code files present in the book (Chapter 1-17).
### Software and Hardware List
| Chapter | Software required | OS required |
| -------- | ------------------------------------ | ----------------------------------- |
| 3-16 | Uflex, Jflex, Unicon, Java | Windows, Mac OS X, and Linux (Any) |

We also provide a PDF file that has color images of the screenshots/diagrams used in this book. [Click here to download it](https://static.packt-cdn.com/downloads/9781800204805_ColorImages.pdf).

## Code in Action
The Code in Action videos for this book can be viewed at https://bit.ly/3njc15D.

### Related products
* Supercharge Your Applications with GraalVM  [[Packt]](https://www.packtpub.com/product/supercharge-your-applications-with-graalvm/9781800564909?utm_source=github&utm_medium=repository&utm_campaign=9781800564909) [[Amazon]](https://www.amazon.com/dp/1800564902)

* Learn LLVM 12  [[Packt]](https://www.packtpub.com/product/learn-llvm-12/9781839213502?utm_source=github&utm_medium=repository&utm_campaign=9781839213502) [[Amazon]](https://www.amazon.com/dp/1839213507)

* LLVM Techniques, Tips, and Best Practices Clang and Middle-End Libraries  [[Packt]](https://www.packtpub.com/product/llvm-techniques-tips-and-best-practices-clang-and-middle-end-libraries/9781838824952?utm_source=github&utm_medium=repository&utm_campaign=9781838824952) [[Amazon]](https://www.amazon.com/dp/1838824952)


## Get to Know the Author
**Clinton L. Jeffery**
is Professor and Chair of the Department of Computer Science and Engineering at New Mexico Institute of Mining and Technology. He received his B.S. from the University of Washington, and M.S. and Ph.D. degrees from the University of Arizona, all in computer science. He has conducted research and written many books and papers on programming languages, program monitoring, debugging, graphics, virtual environments, and visualization. With colleagues, he invented the Unicon programming language, hosted on the Unicon website.

